@@ -4,7 +4,7 @@ use warnings FATAL => 'all';
 
 package namespace::local;
 
-our $VERSION = '0.06_02';
+our $VERSION = '0.07_01';
 
 =head1 NAME
 
@@ -156,51 +156,53 @@ work as expected if used simultaneously.
 
 =cut
 
-use Carp;
-
 # this was stolen from namespace::clean
 use B::Hooks::EndOfScope 'on_scope_end';
-use Scalar::Util qw( weaken );
 
-my $last_command;
+my @stack;
 
 sub import {
     my $class = shift;
 
-    my $command = namespace::local::_command->new( target => scalar caller );
+    my $command = namespace::local::_command->new( caller => [ caller ] );
     $command->parse_options( @_ );
 
     # on_scope_end callback execution order is direct
     # we need reversed order, so use a stack of commands.
-    $last_command->set_next( $command ) if $last_command;
-    $last_command = $command;
-    weaken $last_command; # to avoid leaks
+    $stack[-1]->set_next( $command ) if @stack;
+    push @stack, $command;
 
     $command->prepare;
 
     on_scope_end {
+        local $Carp::Internal{'B::Hooks::EndOfScope::XS'} = 1;
+        local $Carp::Internal{'B::Hooks::EndOfScope'} = 1;
+        pop @stack;
         $command->execute;
     };
 };
 
-;
-
 # Hide internal OO engine
 # Maybe it will be released later...
 
 package
     namespace::local::_command;
 
-use Carp; # too
+use Carp;
+use Scalar::Util qw(refaddr);
 our @CARP_NOT = qw(namespace::local);
 
-# target => package_name is required
+### Setup methods
+
+# requires caller => [caller] argument
 sub new {
     my ($class, %opt) = @_;
 
     # TODO check options
-    $opt{except_rex} = qr/^[0-9]+$|^-$/; # no matter what, exempt $_, $1, ...
+    $opt{except_rex} = qr/^[0-9]+$|^_$/; # no matter what, exempt $_, $1, ...
     $opt{action} = '-around';
+    $opt{target} = $opt{caller}[0];
+    $opt{origin} = join ":", @{$opt{caller}}[1,2];
 
     # Skip some well-known variables and functions
     # Format: touch_not{ $name }{ $type }
@@ -215,9 +217,19 @@ sub new {
 
 sub set_next {
     my ($self, $next) = @_;
+
+    carp "probably a bug in namespace::local - uncommitted command replaced in chain"
+        if $self->{next} and !$self->{next}{done};
+
     $self->{next} = $next;
 };
 
+sub DESTROY {
+    my $self = shift;
+    carp "probably a bug in namespace::local: callback set but not executed"
+        if $self->{todo} and !$self->{done};
+};
+
 my %known_action;
 $known_action{$_}++ for qw(-above -below -around);
 
@@ -264,6 +276,8 @@ sub touch_not {
     };
 };
 
+### Command pattern split into prepare + execute
+
 # side effects + setup self->execute
 sub prepare {
     my $self = shift;
@@ -272,14 +286,11 @@ sub prepare {
 
     my $table = $self->read_symbols;
 
-    # FIXME UGLY HACK
-    # Immediate backup-and-restore of symbol table
-    #     somehow forces binding of symbols
-    #     above 'use namespace::local' line
-    #     thus preventing subsequent imports from leaking upwards
-    # I do not know why it works, it shouldn't.
     if ($action eq '-around') {
-        $self->erase_symbols;
+        # Overwrite symbol table with a copy of itself.
+        # Somehow this triggers binding of symbols in the code
+        #    that was parsed so far (i.e. above the 'use' line)
+        #    and undefined symbols (in that area) remain so forever
         $self->write_symbols( $table );
     };
 
@@ -289,8 +300,7 @@ sub prepare {
         };
     } else {
         $self->{todo} = sub {
-            $self->erase_symbols;
-            $self->write_symbols( $table );
+            $self->replace_symbols( undef, $table );
         };
     };
 };
@@ -306,14 +316,124 @@ sub execute {
         unless $self->{done}++;
 };
 
-# in: package name
+### High-level effectful functions
+
+# Here and below, the following data format is used:
+# `symtable` := { `name` => { `type` => `$ref` } }
+# `type` is one of known variable types listed in @TYPES below
+# `$ref` is a reference of corresponding type
+
+# Don't touch NAME, PACKAGE, and GLOB that are alsoknown to Perl
+my @TYPES = qw(SCALAR ARRAY HASH CODE IO FORMAT);
+
+# In: symbol table hashref
+# Out: side effect
+sub erase_only_symbols {
+    my ($self, $table) = @_;
+
+    my $package = $self->{target};
+    my @list = keys %$table;
+
+    # load all necessary symbols
+    my $current = $self->read_symbols( \@list );
+
+    # filter out what we were going to delete
+    foreach my $name ( @list ) {
+        $table->{$name}{$_} and delete $current->{$name}{$_}
+            for @TYPES;
+    };
+
+    # put it back in place
+    $self->replace_symbols( \@list, $current );
+};
+
+# This method's signature is a bit counterintuitive:
+# $self->replace_symbols( \@names_to_erase, \%new_table_entries )
+# If first argument is omitted, the whole namespace is scanned instead.
+# Separate erase_symbols and write_symbols turned out to be cumbersome
+#    because of the need to handle granular exclusion list.
+# This method can fill both roles.
+# Providing an empty list would make it just write the symbol table,
+#    whereas an empty hash would mean deletion only.
+sub replace_symbols {
+    my ($self, $clear_list, $table) = @_;
+
+    $clear_list ||= [ $self->read_names ];
+    $table ||= {};
+
+    my %uniq;
+    $uniq{$_}++ for keys %$table, @$clear_list;
+
+    # re-read the symbol table
+    my $old_table = $self->read_symbols( [ keys %uniq ] );
+
+    # create a plan for change
+    my $diff = $self->table_diff( $old_table, $table );
+
+    # apply change
+    $self->write_symbols( $diff );
+};
+
+# Oddly enough, pure
+# In: old and new two symbol table hashrefs
+# Out: part of new table that differs from the old,
+#      with touch_not rules applied
+sub table_diff {
+    my ($self, $old_table, $new_table) = @_;
+
+    my %uniq_name;
+    $uniq_name{$_}++ for keys %$old_table, keys %$new_table;
+
+    my $diff;
+
+    # iterate over keys of both, 2 levels deep
+    foreach my $name (sort keys %uniq_name) {
+        my $old  = $old_table->{$name} || {};
+        my $new  = $new_table->{$name} || {};
+        my $skip = $self->{touch_not}{$name} || {};
+
+        my %uniq_type;
+        $uniq_type{$_}++ for keys %$old, keys %$new;
+
+        foreach my $type (sort keys %uniq_type) {
+            next if $skip->{$type};
+
+            if (ref $old->{$type} ne ref $new->{$type}) {
+                # As nonrefs are not allowed here,
+                # this also handles undef vs. defined case
+                $diff->{$name}{$type} = $new->{$type};
+                next;
+            };
+
+            # both undef, nothing to see here
+            next unless ref $new->{$type};
+
+            # pointing to different things
+            if (refaddr $old->{$type} != refaddr $new->{$type}) {
+                $diff->{$name}{$type} = $new->{$type};
+                next;
+            };
+        };
+
+        if ($diff->{$name}) {
+            # now if we cannot avoid overwriting,
+            # make sure to copy over ALL skipped values for this name
+            $diff->{$name}{$_} = $old->{$_} for keys %$skip;
+        };
+    };
+
+    return $diff;
+};
+
+### Low-level symbol-table read & write
+### no magic should happen above this line
+
+# NOTE that even here we are in full strict mode
+# The pattern for working with raw data is this:
+# my $value = do { no strict 'refs'; ... }; ## no critic
+
+# in: none
 # out: sorted & filtered list of symbols
-
-# FIXME needs explanation
-# We really need to filter because copying ALL table
-#     was preventing on_scope_end from execution
-#     (accedental reference count increase?..)
-
 sub read_names {
     my $self = shift;
 
@@ -330,45 +450,8 @@ sub read_names {
     return @list;
 };
 
-# Don't touch NAME, PACKAGE, and GLOB itself
-my @TYPES = qw(SCALAR ARRAY HASH CODE IO FORMAT);
-
-# In: package
-# Out: (none)
-# Side effect: destroys symbol table
-sub erase_symbols {
-    my ($self, $list) = @_;
-
-    my $package = $self->{target};
-    $list ||= [ $self->read_names ];
-
-    foreach my $name( @$list ) {
-        next if $self->{touch_not}{$name};
-        no strict 'refs'; ## no critic
-        delete ${ $package."::" }{$name};
-    };
-};
-
-sub erase_only_symbols {
-    my ($self, $table) = @_;
-
-    my $package = $self->{target};
-    my @list = keys %$table;
-
-    # read all necessary symbols
-    my $current = $self->read_symbols( \@list );
-
-    foreach my $name ( @list ) {
-        $table->{$name}{$_} and delete $current->{$name}{$_}
-            for @TYPES;
-    };
-
-    $self->erase_symbols( \@list );
-    $self->write_symbols( $current );
-};
-
-# In: package, symbol
-# Out: a hash with glob content
+# In: symbol list arrayref (read_symbols if none)
+# Out: symbol table hashref
 sub read_symbols {
     my ($self, $list) = @_;
 
@@ -389,41 +472,35 @@ sub read_symbols {
     return \%content;
 };
 
-# In: package, symbol, hash
-# Out: (none)
-# Side effect: recreates *package::symbol
+# writes raw symbols, ignoring touch_not!
+# In: symbol table hashref
+# Out: none
 sub write_symbols {
     my ($self, $table) = @_;
 
     my $package = $self->{target};
 
-    # TODO lumping touch_not and $table together is a workaround
-    #     for erase_symbols skipping over touch_not altogether.
-    # It does not belong here, and should be fixed at some point.
-    my %uniq;
-    $uniq{$_}++ for keys %$table, keys %{ $self->{touch_not} };
-
-    foreach my $name( keys %uniq ) {
-        my $copy = $table->{$name} || {};
-        if ( my $skip = $self->{touch_not}{$name} ) {
-            foreach my $type (keys %$skip) {
-                my $value = do {
-                    no strict 'refs'; ## no critic
-                    *{$package."::".$name}{$type};
-                };
-                $copy->{$type} = $value if defined $value;
-            };
-        };
+    foreach my $name( keys %$table ) {
+        my $copy = $table->{$name};
 
         {
             no strict 'refs'; ## no critic
             delete ${ $package."::" }{$name};
         };
 
-        foreach my $type ( @TYPES ) {
-            defined $copy->{$type} or next;
-            no strict 'refs'; ## no critic
-            *{ $package."::".$name } = $copy->{$type}
+        foreach my $type ( keys %$copy ) {
+            ref $copy->{$type} or next;
+            eval {
+                # FIXME on perls 5.014..5.022 this block fails
+                # because @ISA is readonly.
+                # So we wrap it in eval with no catch
+                # until a better solution is done
+                no strict 'refs'; ## no critic
+                *{ $package."::".$name } = $copy->{$type};
+                1;
+            } || do {
+                carp "namespace::local: failed to write $package :: $name ($type), but trying to continue: $@";
+            };
         };
     };
 };
@@ -432,13 +509,13 @@ sub _croak {
     croak ("namespace::local: ".shift);
 };
 
-=head1 AUTHOR
-
-Konstantin S. Uvarin, C<< <khedin at gmail.com> >>
-
 =head1 BUGS
 
-This is experimental module. There certainly are bugs.
+As of 0.0604, C<-around> hides subroutines defined below its scope end
+from anything above it.
+No solution exists so far.
+
+This is experimental module. There certainly are more bugs.
 
 Bug reports, feature requests, suggestions and general feedback welcome at:
 
@@ -492,7 +569,7 @@ L<B::Hooks::EndOfScope> is used as a backend.
 
 =head1 LICENSE AND COPYRIGHT
 
-Copyright 2018 Konstantin S. Uvarin.
+Copyright 2018 Konstantin S. Uvarin, C<< <khedin at gmail.com> >>
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the the Artistic License (2.0). You may obtain a
